<!DOCTYPE html>
<html lang="{{ .Site.Language.Lang }}" dir="{{ .Site.Language.LanguageDirection | default `ltr` }}">
  {{- partial "head.html" . -}}
  <body>
    <a href="#content" class="hx:sr-only hx:focus-visible:not-sr-only hx:focus-visible:fixed hx:focus-visible:z-50 hx:focus-visible:top-2 hx:focus-visible:left-2 hx:focus-visible:bg-primary-500 hx:focus-visible:text-white hx:focus-visible:px-4 hx:focus-visible:py-2 hx:focus-visible:rounded-md hx:focus-visible:text-sm hx:focus-visible:font-medium">
      {{- (T "skipToContent") | default "Skip to content" -}}
    </a>
    {{- partial "banner.html" . -}}
    {{- partial "navbar.html" . -}}
    {{- block "main" . }}{{ end -}}
    {{- if or (eq .Site.Params.footer.enable nil) (.Site.Params.footer.enable) }}
      {{ partial "footer.html" . }}
    {{ end }}
    {{ partial "scripts.html" . }}
  </body>
</html>
