<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Page Not Found</title>
    <style>
        body { text-align: center; font-family: sans-serif; padding: 100px 20px; background: #f9f9f9; color: #333; }
        h1 { font-size: 48px; margin-bottom: 10px; }
        p { font-size: 18px; color: #666; margin-bottom: 30px; }
        a { display: inline-block; padding: 10px 20px; background: #0056b3; color: white; text-decoration: none; border-radius: 5px; }
        a:hover { background: #004085; }
    </style>
</head>
<body>
    <h1>404</h1>
    <p>Sorry, the page you are looking for doesn't exist.</p>
    <a href="{{ site.BaseURL | default "/" }}">Return to Homepage</a>
</body>
</html>
