<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="UTF-8">
  <meta http-equiv="refresh" content="0; url={{ .Site.BaseURL }}docs/">
  <link rel="canonical" href="{{ .Site.BaseURL }}docs/">
  <script>window.location.replace("{{ .Site.BaseURL }}docs/");</script>
</head>
<body>
  <a href="{{ .Site.BaseURL }}docs/">Redirecting…</a>
</body>
</html>