///\file

/******************************************************************************
The MIT License(MIT)

Embedded Template Library.
https://github.com/ETLCPP/etl
https://www.etlcpp.com

Copyright(c) 2020 John Wellbelove

Permission is hereby granted" free of charge" to any person obtaining a copy
of this software and associated documentation files(the "Software")" to deal
in the Software without restriction" including without limitation the rights
to use" copy" modify" merge" publish" distribute" sublicense" and / or sell
copies of the Software" and to permit persons to whom the Software is
furnished to do so" subject to the following conditions :

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS"" WITHOUT WARRANTY OF ANY KIND" EXPRESS OR
IMPLIED" INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY"
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM" DAMAGES OR OTHER
LIABILITY" WHETHER IN AN ACTION OF CONTRACT" TORT OR OTHERWISE" ARISING FROM"
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
******************************************************************************/

#ifndef ETL_FILE_ERROR_NUMBERS
#define ETL_FILE_ERROR_NUMBERS

#define ETL_DEQUE_FILE_ID                          "1"
#define ETL_FLAT_MAP_FILE_ID                       "2"
#define ETL_FLAT_MULTIMAP_FILE_ID                  "3"
#define ETL_FLAT_MULTISET_FILE_ID                  "4"
#define ETL_FLAT_SET_FILE_ID                       "5"
#define ETL_FORWARD_LIST_FILE_ID                   "6"
#define ETL_LIST_FILE_ID                           "7"
#define ETL_MAP_FILE_ID                            "8"
#define ETL_MULTIMAP_FILE_ID                       "9"
#define ETL_MULTISET_FILE_ID                       "10"
#define ETL_POOL_FILE_ID                           "11"
#define ETL_PRIORITY_QUEUE_FILE_ID                 "12"
#define ETL_QUEUE_FILE_ID                          "13"
#define ETL_SET_FILE_ID                            "14"
#define ETL_STACK_FILE_ID                          "15"
#define ETL_UNORDERED_MAP_FILE_ID                  "16"
#define ETL_VECTOR_FILE_ID                         "17"
#define ETL_OBSERVER_FILE_ID                       "18"
#define ETL_IHASH_FILE_ID                          "19"
#define ETL_INTRUSIVE_FORWARD_LIST_FILE_ID         "20"
#define ETL_INTRUSIVE_LIST_FILE_ID                 "21"
#define ETL_INTRUSIVE_LINKS_FILE_ID                "22"
#define ETL_UNORDERED_SET_FILE_ID                  "23"
#define ETL_VARIANT_FILE_ID                        "24"
#define ETL_UNORDERED_MULTIMAP_FILE_ID             "25"
#define ETL_UNORDERED_MULTISET_FILE_ID             "26"
#define ETL_BASIC_STRING_FILE_ID                   "27"
#define ETL_INTRUSIVE_STACK_FILE_ID                "28"
#define ETL_INTRUSIVE_QUEUE_FILE_ID                "29"
#define ETL_REFERENCE_FLAT_MAP_FILE_ID             "30"
#define ETL_REFERENCE_FLAT_MULTIMAP_FILE_ID        "31"
#define ETL_REFERENCE_FLAT_SET_FILE_ID             "32"
#define ETL_REFERENCE_FLAT_MULTISET_FILE_ID        "33"
#define ETL_FSM_FILE_ID                            "34"
#define ETL_MESSAGE_ROUTER_FILE_ID                 "35"
#define ETL_SCHEDULER_FILE_ID                      "36"
#define ETL_TASK_FILE_ID                           "37"
#define ETL_MESSAGE_FILE_ID                        "38"
#define ETL_MESSAGE_BUS_FILE_ID                    "39"
#define ETL_VARIANT_POOL_FILE_ID                   "40"
#define ETL_ARRAY_VIEW_FILE_ID                     "41"
#define ETL_STRING_VIEW_FILE_ID                    "42"
#define ETL_CALLBACK_TIMER_FILE_ID                 "43"
#define ETL_MESSAGE_TIMER_FILE_ID                  "44"
#define ETL_TYPE_LOOKUP_FILE_ID                    "45"
#define ETL_QUEUE_SPSC_ISR_FILE_ID                 "46"
#define ETL_QUEUE_SPSC_ATOMIC_FILE_ID              "47"
#define ETL_QUEUE_MPMC_MUTEX_FILE_ID               "48"
#define ETL_TYPE_SELECT_FILE_ID                    "49"
#define ETL_BINARY_FILE_ID                         "50"
#define ETL_DELEGATE_FILE_ID                       "51"
#define ETL_BITSET_FILE_ID                         "52"
#define ETL_INDIRECT_VECTOR_FILE_ID                "53"
#define ETL_QUEUE_SPSC_LOCKED_FILE_ID              "54"
#define ETL_MESSAGE_PACKET_FILE_ID                 "55"
#define ETL_CIRCULAR_BUFFER_FILE_ID                "56"
#define ETL_MULTI_LOOP_FILE_ID                     "57"
#define ETL_REFERENCE_COUNTER_MESSAGE_POOL_FILE_ID "58"
#define ETL_QUEUE_SPSC_LOCKABLE_FILE_ID            "59"
#define ETL_MESSAGE_ROUTER_REGISTRY_FILE_ID        "60"
#define ETL_ARRAY_WRAPPER_FILE_ID                  "61"
#define ETL_MEM_CAST_FILE_ID                       "62"
#define ETL_SINGLETON_FILE_ID                      "63"
#define ETL_SUCCESSOR_FILE_ID                      "64"
#define ETL_BIT_STREAM_FILE_ID                     "65"
#define ETL_BYTE_STREAM_FILE_ID                    "66"
#define ETL_BIP_BUFFER_SPSC_ATOMIC_FILE_ID         "67"
#define ETL_REFERENCE_COUNTED_OBJECT_FILE_ID       "68"
#define ETL_TO_ARITHMETIC_FILE_ID                  "69"
#define ETL_EXPECTED_FILE_ID                       "70"
#define ETL_ALIGNMENT_FILE_ID                      "71"
#define ETL_BASE64_FILE_ID                         "72"
#define ETL_SINGLETON_BASE_FILE_ID                 "73"
#define ETL_UNALIGNED_TYPE_FILE_ID                 "74"
#define ETL_SPAN_FILE_ID                           "75"
#define ETL_ALGORITHM_FILE_ID                      "76"
#define ETL_NOT_NULL_FILE_ID                       "77"
#define ETL_SIGNAL_FILE_ID                         "78"
#define ETL_FORMAT_FILE_ID                         "79"
#define ETL_INPLACE_FUNCTION_FILE_ID               "80"
#define ETL_INTRUSIVE_AVL_TREE_FILE_ID             "81"
#endif
