/******************************************************************************
The MIT License(MIT)

Embedded Template Library.
https://github.com/ETLCPP/etl
https://www.etlcpp.com

Copyright(c) 2021 Bo Rydberg

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files(the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions :

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
******************************************************************************/

#ifndef __ETL_PROFILE_H__
#define __ETL_PROFILE_H__

#define ETL_TARGET_DEVICE_GENERIC
#define ETL_TARGET_OS_NONE
#define ETL_CALLBACK_TIMER_USE_INTERRUPT_LOCK
#define ETL_CALLBACK_TIMER_DISABLE_INTERRUPTS ((void)0)
#define ETL_CALLBACK_TIMER_ENABLE_INTERRUPTS  ((void)0)
#define ETL_MESSAGE_TIMER_USE_INTERRUPT_LOCK
#define ETL_MESSAGE_TIMER_DISABLE_INTERRUPTS ((void)0)
#define ETL_MESSAGE_TIMER_ENABLE_INTERRUPTS  ((void)0)

#if defined(ETL_FORCE_TEST_CPP03_IMPLEMENTATION)
  #define ETL_FUNCTION_FORCE_CPP03_IMPLEMENTATION
  #define ETL_PRIORITY_QUEUE_FORCE_CPP03_IMPLEMENTATION
  #define ETL_QUEUE_ATOMIC_FORCE_CPP03_IMPLEMENTATION
  #define ETL_VARIANT_FORCE_CPP03_IMPLEMENTATION
  #define ETL_VECTOR_FORCE_CPP03_IMPLEMENTATION
  #define ETL_QUEUE_FORCE_CPP03_IMPLEMENTATION
  #define ETL_QUEUE_MPMC_MUTEX_FORCE_CPP03_IMPLEMENTATION
  #define ETL_QUEUE_ISR_FORCE_CPP03_IMPLEMENTATION
  #define ETL_QUEUE_LOCKED_FORCE_CPP03_IMPLEMENTATION
  #define ETL_OPTIONAL_FORCE_CPP03_IMPLEMENTATION
  #define ETL_LARGEST_TYPE_FORCE_CPP03_IMPLEMENTATION
  #define ETL_TYPE_SELECT_FORCE_CPP03_IMPLEMENTATION
  #define ETL_UNINITIALIZED_BUFFER_FORCE_CPP03_IMPLEMENTATION
  #define ETL_CRC_FORCE_CPP03_IMPLEMENTATION
  #define ETL_MEM_CAST_FORCE_CPP03_IMPLEMENTATION
  #define ETL_OBSERVER_FORCE_CPP03_IMPLEMENTATION
  #define ETL_MESSAGE_PACKET_FORCE_CPP03_IMPLEMENTATION
  #define ETL_OBSERVER_FORCE_CPP03_IMPLEMENTATION
  #define ETL_MESSAGE_ROUTER_FORCE_CPP03_IMPLEMENTATION
  #define ETL_FSM_FORCE_CPP03_IMPLEMENTATION
  #define ETL_SINGLETON_FORCE_CPP03_IMPLEMENTATION
#endif

#endif
